from logging import Manager
from os import makedirs
from os import path as pt

import torch

from neuralhedge.data.base import HedgerDataset, ManagerDataset
from neuralhedge.nn import datahedger, loss, network
from neuralhedge.nn.loss import (
    EntropicRiskMeasure,
    ExpectedShortfall,
    PowerMeasure,
    SquareMeasure,
    log_utility,
)
from neuralhedge.utils.plotting import plot_hedge, plot_history, plot_pnl, plot_strategy


def evaluate_bs_deep_hedge(
    hedger: datahedger.Hedger, ds: HedgerDataset, bs_price, record_dir: str = ""
):
    if record_dir:
        print(f"Save evaluation at: {record_dir}")
        makedirs(record_dir, exist_ok=True)
        torch.save(hedger.strategy.state_dict(), pt.join(record_dir, "model.pth"))

    init_wealth = bs_price
    prices, info, payoff = ds.data
    with torch.no_grad():
        wealth = hedger.forward(prices, info, init_wealth)
        terminal_wealth = wealth[:, -1]
        pnl = terminal_wealth - payoff

    print(f"Entropic Measure: {EntropicRiskMeasure()(pnl):.2f}")
    print(f"Square Measure: {SquareMeasure()(pnl):.2f}")
    print(f"Power Measure (p=1): {PowerMeasure(p=1)(pnl):.2f}")
    print(f"Expected Shortfall (q=0.5): {ExpectedShortfall(q=0.5)(pnl):.2f}")
    print(f"Expected Shortfall (q=0.9): {ExpectedShortfall(q=0.9)(pnl):.2f}")

    plot_pnl(pnl, record_dir)
    plot_hedge(prices, terminal_wealth, payoff, init_wealth, record_dir)
    plot_strategy(hedger.strategy, ds, record_dir)


def evaluate_bs_efficient_hedge(
    hedger: datahedger.Hedger,
    ds: HedgerDataset,
    bs_price,
    init_wealth,
    record_dir: str = "",
):
    if record_dir:
        print(f"Save evaluation at: {record_dir}")
        makedirs(record_dir, exist_ok=True)
        torch.save(hedger.strategy.state_dict(), pt.join(record_dir, "model.pth"))

    print(f"BS price: {bs_price:.2f}")
    print(f"initial wealth: {init_wealth:.2f}")

    prices, info, payoff = ds.data
    with torch.no_grad():
        wealth = hedger.forward(prices, info, init_wealth)
        terminal_wealth = wealth[:, -1]
        pnl = terminal_wealth - payoff

    print(f"Risk(p=1): {loss.PowerMeasure(1)(pnl):.2f}")
    print(f"Risk(p=2): {loss.PowerMeasure(2)(pnl):.2f}")
    print(f"Risk(p=0.5): {loss.PowerMeasure(0.5)(pnl):.2f}")

    print(f"Entropic Measure: {EntropicRiskMeasure()(pnl):.2f}")
    print(f"Square Measure: {SquareMeasure()(pnl):.2f}")
    print(f"Expected Shortfall (q=0.5): {ExpectedShortfall(q=0.5)(pnl):.2f}")
    print(f"Expected Shortfall (q=0.9): {ExpectedShortfall(q=0.9)(pnl):.2f}")

    plot_pnl(pnl, record_dir)
    plot_hedge(prices, terminal_wealth, payoff, init_wealth, record_dir)
    plot_strategy(hedger.strategy, ds, record_dir)


from neuralhedge.nn.datamanager import Manager


def evaluate_merton(
    hedger: Manager,
    ds: ManagerDataset,
    init_wealth,
    record_dir: str = "",
):
    if record_dir:
        print(f"Save evaluation at: {record_dir}")
        makedirs(record_dir, exist_ok=True)
        torch.save(hedger.strategy.state_dict(), pt.join(record_dir, "model.pth"))

    print(f"initial wealth: {init_wealth:.2f}")

    prices, info = ds.data
    with torch.no_grad():
        wealth = hedger.forward(prices, info)
        terminal_wealth = wealth[:, -1] * init_wealth

    utility = log_utility(terminal_wealth).mean().numpy()
    print(f"Log utility: {utility:.2f}")

    plot_pnl(terminal_wealth, record_dir)


def evaluate_mv(
    hedger: Manager,
    ds: ManagerDataset,
):
    prices, info = ds.data
    with torch.no_grad():
        wealth = hedger.forward(prices, info)
        terminal_wealth = wealth[:, -1]
    mean = terminal_wealth.mean().item()
    var = terminal_wealth.var().item()
    return mean, var



import torch

from neuralhedge.data.base import HedgerDataset, ManagerDataset
from neuralhedge.data.stochastic import BlackScholes, simulate_time
from neuralhedge.nn import blackschole, datahedger
from neuralhedge.nn.contigent import EuropeanVanilla


class BS_Market:
    """
    Data of BS stock + Bond + European call option
    For Portfolio Management, the option part is redundant
    """

    def __init__(
        self,
        n_sample=10000,
        n_timestep=30,
        dt=1 / 30,
        mu=0.1,
        sigma=0.2,
        r=0.0,
        init_price=100.0,
    ) -> None:
        self.n_sample = n_sample
        self.n_timestep = n_timestep
        self.dt = dt
        self.mu = mu
        self.sigma = sigma
        self.r = r
        self.T = n_timestep * self.dt
        self.bs = BlackScholes(
            n_sample=n_sample, n_timestep=n_timestep, dt=dt, mu=mu, sigma=sigma
        )

        self.init_price = init_price

        self.strike = 100.0
        self.contigent = EuropeanVanilla(strike=self.strike, call=True)

        bs_pricer = blackschole.BlackScholesPrice(
            sigma=sigma, risk_free_rate=r, strike=self.contigent.strike
        )

        tmp = torch.tensor([torch.tensor(self.init_price).log(), self.T])
        self.bs_price = bs_pricer(tmp)[0]
        self.bs_delta = blackschole.BlackScholesDelta(
            sigma=sigma, risk_free_rate=r, strike=self.contigent.strike
        )

    def get_hedge_ds(self):
        stock_prices = self.bs.get_prices() * self.init_price
        time = simulate_time(self.n_sample, self.dt, self.n_timestep, reverse=False)
        bond_prices = torch.exp(self.r * time) * self.init_price
        prices = torch.cat([stock_prices, bond_prices], dim=-1)
        payoff = self.contigent.payoff(prices[:, -1, 0])
        info = torch.cat(
            [
                torch.log(prices[..., :1]),
                torch.sqrt(
                    simulate_time(self.n_sample, self.dt, self.n_timestep, reverse=True)
                ),
            ],
            dim=-1,
        )
        data = (prices, info, payoff)
        hedge_ds = HedgerDataset(*data)
        return hedge_ds

    def get_manage_ds(self):
        stock_prices = self.bs.get_prices() * self.init_price
        time = simulate_time(self.n_sample, self.dt, self.n_timestep, reverse=False)
        bond_prices = torch.exp(self.r * time) * self.init_price
        prices = torch.cat([stock_prices, bond_prices], dim=-1)
        # TODO: info in dict form, now 1st log price, second time to maturity
        info = torch.cat(
            [
                torch.log(prices[..., :1]),
                simulate_time(self.n_sample, self.dt, self.n_timestep, reverse=True),
            ],
            dim=-1,
        )
        data = (prices, info)
        manage_ds = ManagerDataset(*data)
        return manage_ds

    def get_price_delta(self):
        return self.bs_price, self.bs_delta


# def stock_bond_market(mu = 0.1,
#                       sigma = 0.2,
#                       r = 0.01,
#                       n_sample = 5000,
#                       n_timestep = 60,
#                       dt = 1/12):
#     # Stock
#     bs = BlackScholes(n_sample = n_sample,
#                         n_timestep = n_timestep,
#                         dt = dt,
#                         mu = mu,
#                         sigma = sigma)
#     # Bond
#     time = simulate_time(n_sample, dt, n_timestep, reverse = False)
#     bond = torch.exp(r*time)
#     # Prices
#     prices = torch.cat([bs.prices, bond], dim = -1)
#     # Information
#     info1 = torch.log(prices)
#     info2 = simulate_time(n_sample, dt, n_timestep, reverse = True)
#     info = torch.cat([info1,
#                 info2],
#                 dim = -1)
#     return prices, info



from typing import List, Tuple
import torch
from torch import Tensor
from torch.utils.data import Dataset


class HedgerDataset(Dataset):
    """Market information dataset.
    Args:
        - prices, information, payoff
    Shape:
        - prices: (n_samples, n_steps+1, n_assets)
        - information: (n_samples, >= n_steps , n_features)
        - payoff: (n_samples,)
    """

    def __init__(self, prices: Tensor, info: Tensor, payoff: Tensor):
        self.data = (prices, info, payoff)
        self.prices = prices
        self.info = info
        self.payoff = payoff
        # TODO: check shape
        # TODO: check input type

    def __len__(self):
        return len(self.prices)

    def __getitem__(self, idx: int):
        return [self.prices[idx], self.info[idx], self.payoff[idx]]


class ManagerDataset(Dataset):
    """Market information dataset.
    Args:
        - prices, information
    Shape:
        - prices: (n_samples, n_steps+1, n_assets)
        - information: (n_samples, >= n_steps , n_features)
    """

    def __init__(
        self,
        prices: Tensor,
        info: Tensor,
    ):
        self.data = (prices, info)
        self.prices = prices
        self.info = info

    def __len__(self):
        return len(self.prices)

    def __getitem__(self, idx: int):
        return [self.prices[idx], self.info[idx]]



from os import path as pt

import matplotlib.pyplot as plt
import numpy as np
import torch
from torch import Tensor

from neuralhedge.nn.loss import EntropicRiskMeasure, SquareMeasure


def to_numpy(x: torch.Tensor):
    return x.numpy()


def plot_hedge(prices, terminal_wealth, payoff, init_wealth, record_dir=None):

    fig = plt.figure()
    plt.scatter(prices[:, -1, 0], terminal_wealth)
    plt.scatter(prices[:, -1, 0], payoff)

    plt.title(f"Hedge (inital:{init_wealth.numpy():.2f})")
    plt.xlabel("Stock")
    plt.ylabel("Wealth")
    plt.grid()

    if record_dir:
        file_path = pt.join(record_dir, "plot_hedge_payoff.png")
        plt.savefig(file_path)
    else:
        plt.show()
    plt.close(fig)


def plot_pnl(pnl, record_dir=None):
    fig = plt.figure()
    plt.hist(pnl, bins=100)
    plt.title("Profit-Loss Histograms")
    plt.xlabel("Profit-Loss")
    plt.ylabel("Number of events")
    plt.grid()
    if record_dir:
        file_path = pt.join(record_dir, "plot_pnl_histogram.png")
        plt.savefig(file_path)
    else:
        plt.show()
    plt.close(fig)


def plot_history(history, record_dir=None):
    smooth_history = np.convolve(np.array(history["loss"]), np.ones(100) / 100, "valid")
    # smooth_history = history
    fig, ax = plt.subplots()
    plt.plot(smooth_history)
    plt.xlabel("Number of epochs")
    plt.ylabel("Loss")
    plt.title("Loss trajectory ")
    if record_dir:
        file_path = pt.join(record_dir, "plot_train_loss.png")
        fig.savefig(file_path)
    else:
        plt.show()
    plt.close(fig)


def plot_hedge_ds(ds, plot_samples=100):
    paths, info, payoff = ds.data
    print("Shape of paths: ", paths.shape)
    print("Shape of information: ", info.shape)
    print("Shape of payoff: ", payoff.shape)

    plt.figure
    plt.plot(to_numpy(paths)[:plot_samples, :, 0].T)
    plt.plot(to_numpy(paths)[0, :, -1], label="risk-free", c="k")
    plt.title("prices paths")
    plt.legend()
    plt.grid()
    plt.show()

    plt.figure
    plt.scatter(to_numpy(paths)[:, -1, 0], to_numpy(payoff))
    plt.title("payoff")
    plt.grid()
    plt.show()


def plot_manage_ds(ds, plot_samples=100):
    paths, info = ds.data
    print("Shape of paths: ", paths.shape)
    print("Shape of information: ", info.shape)

    plt.figure
    plt.plot(to_numpy(paths)[:plot_samples, :, 0].T)
    plt.plot(to_numpy(paths)[0, :, -1], label="risk-free", c="k")
    plt.title("prices paths")
    plt.legend()
    plt.grid()
    plt.show()


def plot_strategy(strategy, ds_test, record_dir=""):
    observe_window = 3
    n_timestep = ds_test.prices.shape[1]
    observe_time = np.unique(np.arange(n_timestep) // observe_window) * observe_window

    fig, axes = plt.subplots(
        3,
        len(observe_time) // 3 + 1,
        figsize=[12, 5],
        sharex=True,
        sharey=True,
    )
    for i, t in enumerate(observe_time):
        ax = axes.flatten()[i]  # type: ignore
        with torch.no_grad():
            x = ds_test.prices[:, t, 0]
            y = strategy(ds_test.info[:, t, :])[..., 0]
        ax.set_title(f"t = {t:.2f}")
        ax.scatter(x, y, s=1)
        ax.set_xticklabels([])

    if record_dir:
        file_path = pt.join(record_dir, "strategy.png")
        fig.savefig(file_path)
    else:
        plt.show()
    plt.close(fig)



import torch
from neuralhedge.data.base import HedgerDataset, ManagerDataset
from neuralhedge.data.market import BS_Market

torch.manual_seed(0)


def test_bs_market_hedge():
    bs_market = BS_Market()
    ds = bs_market.get_hedge_ds()
    assert isinstance(ds, HedgerDataset)
    assert (ds.prices, ds.info, ds.payoff) == ds.data
    assert len(ds.prices.shape) == 3
    assert len(ds.info.shape) == 3
    assert len(ds.payoff.shape) == 1
    assert len(ds.prices) == len(ds.info) == len(ds.payoff)


def test_bs_market_manage():
    bs_market = BS_Market()
    ds = bs_market.get_manage_ds()
    assert isinstance(ds, ManagerDataset)
    assert (ds.prices, ds.info) == ds.data
    assert len(ds.prices.shape) == 3
    assert len(ds.info.shape) == 3
    assert len(ds.prices) == len(ds.info)



from typing import List

import torch
from torch import Tensor

from neuralhedge.nn.base import BaseModel
from neuralhedge.nn.loss import EntropicRiskMeasure, LossMeasure, admissible_cost


class Hedger(BaseModel):
    """Hedger to hedge with only data generated but not the generating class
    Args:
        - strategy (torch.nn.Module) or models (List[Module]): depending on independent neural network at each time step or the same neural network at each time
        - risk (HedgeLoss)
    """

    def __init__(
        self, strategy: torch.nn.Module, risk: LossMeasure = EntropicRiskMeasure()
    ):

        super().__init__()
        self.strategy = strategy
        self.risk = risk

    def forward(self, prices: Tensor, info: Tensor, init_wealth: Tensor):

        wealth0_dis_list = self.compute_wealth0_dis(prices, info)
        wealth0_dis = torch.cat(
            [wealth_t.unsqueeze(1) for wealth_t in wealth0_dis_list], dim=1
        )

        init_wealth_dis = init_wealth / prices[:, 0, -1]
        wealth_dis = wealth0_dis + init_wealth_dis.unsqueeze(1)
        wealth = wealth_dis * prices[:, :, -1]
        return wealth

    def compute_wealth0_dis(self, prices, info):
        """
        Compute the discounted wealth process
        """
        batch_size, n_timestep, n_asset = prices.shape
        wealth_dis = [torch.zeros([batch_size]) for t in range(n_timestep)]
        prices_dis = prices[..., :-1] / prices[..., -1:]
        holding_stock = torch.zeros_like(prices_dis)  # t = 0 is meaningless
        for t in range(n_timestep - 1):
            info_dyn = holding_stock
            all_info_t = self.compute_info_t(info_dyn, info, t)
            holding_stock[:, t + 1, :] = self.compute_holding_stock_tplus1(all_info_t)
            wealth_incr = holding_stock[:, t + 1, :] * (
                prices_dis[:, t + 1, :] - prices_dis[:, t, :]
            )
            wealth_incr_sum = torch.sum(wealth_incr, dim=-1, keepdim=False)
            wealth_dis[t + 1] = wealth_dis[t] + wealth_incr_sum
        return wealth_dis

    def compute_info_t(self, info_dyn: Tensor, info: Tensor, t=None) -> Tensor:
        # all_info_t = torch.cat(
        #         [info[:, t, :], info_dyn[:,t,:]],
        #         dim=-1)
        all_info_t = info[:, t, :]
        return all_info_t

    def compute_holding_stock_tplus1(
        self, all_info_t: Tensor, t=None
    ) -> Tensor:  # We might use t here if it is deep hedge
        holding_stock_tplus1 = self.strategy(all_info_t)
        return holding_stock_tplus1

    def compute_pnl(
        self, prices: Tensor, info: Tensor, init_wealth: Tensor, payoff: Tensor
    ):
        wealth = self.forward(prices, info, init_wealth)
        terminal_wealth = wealth[:, -1]
        pnl = terminal_wealth - payoff
        return pnl

    def compute_loss(self, input: List[Tensor]):
        prices, info, payoff = input
        init_wealth = torch.tensor(0.0)
        pnl = self.compute_pnl(prices, info, init_wealth, payoff)
        loss = self.risk(pnl)
        return loss

    def pricer(self, input):
        prices, info, payoff = input
        with torch.no_grad():
            init_wealth = torch.tensor(0.0)
            pnl = self.compute_pnl(prices, info, init_wealth, payoff)
            price = self.risk.cash(pnl)
        return price


class EfficientHedger(Hedger):
    def __init__(
        self,
        strategy: torch.nn.Module,
        init_wealth: Tensor,
        risk: LossMeasure = EntropicRiskMeasure(),
        ad_bound=0.0,
    ):
        super().__init__(strategy, risk)
        self.init_wealth = init_wealth
        self.ad_bound = ad_bound

    def compute_loss(self, input: List[Tensor]):
        prices, info, payoff = input
        init_wealth = self.init_wealth
        wealth = self.forward(prices, info, init_wealth)
        terminal_wealth = wealth[:, -1]
        pnl = terminal_wealth - payoff
        ad_cost = admissible_cost(wealth, self.ad_bound)
        loss = self.risk(pnl) + ad_cost
        return loss



from typing import List
import torch
from torch import Tensor


class BaseModel(torch.nn.Module):

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def compute_loss(self, input: List[Tensor]) -> Tensor:
        return torch.zeros(1)



from collections import defaultdict
from typing import List

import torch
from torch import Tensor
from torch.utils.data import DataLoader, Dataset
from tqdm import tqdm

from neuralhedge.nn.base import BaseModel


class Trainer(torch.nn.Module):
    def __init__(self, model: BaseModel) -> None:
        super().__init__()
        self.model = model
        self.history = defaultdict(list)

    def forward(self, input: List[Tensor]):
        loss = self.model.compute_loss(input)
        return loss

    def fit(
        self,
        hedger_ds: Dataset,
        EPOCHS=100,
        batch_size=256,
        optimizer=torch.optim.Adam,
        lr_scheduler_gamma=1.0,
        lr=0.01,
    ):
        self.steps = 1

        hedger_dl = DataLoader(
            hedger_ds, batch_size=batch_size, shuffle=True, num_workers=0
        )

        self.optimizer = optimizer(self.parameters(), lr=lr)
        lr_scheduler = torch.optim.lr_scheduler.ExponentialLR(
            self.optimizer, gamma=lr_scheduler_gamma
        )

        self.train(True)
        best_loss = 1e10
        progress = tqdm(range(EPOCHS))
        for epoch in progress:
            for i, data in enumerate(hedger_dl):
                self.optimizer.zero_grad()
                loss = self(data)
                self.history["loss"].append(loss.item())
                progress.desc = "Loss=" + str(loss.item())
                if loss.item() < best_loss:
                    best_loss = loss.item()
                    self.best_weights = self.model.state_dict()
                loss.backward()
                self.optimizer.step()
                self.steps += 1
            if epoch % 100 == 0:
                lr_scheduler.step()
        self.model.load_state_dict(self.best_weights)



import torch
from torch import Tensor
from torch.utils.data import Dataset, DataLoader, TensorDataset


def simulate_BM(n_sample, dt, n_timestep):
    noise = torch.randn(size=(n_sample, n_timestep))
    paths_incr = noise * torch.sqrt(torch.tensor(dt))
    paths = torch.cumsum(paths_incr, dim=1)
    BM_paths = torch.cat([torch.zeros((n_sample, 1)), paths], dim=1)
    BM_paths = BM_paths[..., None]
    return BM_paths


def simulate_BS(
    n_sample, dt, n_timestep, mu, sigma
):  # Maybe define it as class method?
    time_grid = torch.linspace(0, dt * n_timestep, n_timestep + 1)
    time_paths = time_grid.expand([n_sample, n_timestep + 1])[..., None]
    BM_paths = simulate_BM(n_sample, dt, n_timestep)
    BS_paths = torch.exp(sigma * BM_paths + (mu - 0.5 * sigma**2) * time_paths)
    return BS_paths


def simulate_time(n_sample, dt, n_timestep, reverse=False):
    time_grid = torch.linspace(0, dt * n_timestep, n_timestep + 1)
    time_paths = time_grid.expand([n_sample, n_timestep + 1])[..., None]
    if reverse:
        return n_timestep * dt - time_paths
    else:
        return time_paths


class BlackScholes:
    def __init__(self, n_sample, n_timestep, dt, mu, sigma):
        self.mu = mu
        self.sigma = sigma
        self.dt = dt
        self.n_sample = n_sample
        self.n_timestep = n_timestep

    def get_prices(self):
        self.prices = simulate_BS(
            self.n_sample, self.dt, self.n_timestep, self.mu, self.sigma
        ).type(torch.float32)
        return self.prices



from copy import deepcopy

import torch
from torch.nn import LazyLinear, Linear, Module, ReLU, Sequential


class NeuralNetSequential(Sequential):
    def __init__(
        self,
        n_output: int = 1,
        n_layers: int = 2,
        n_units: int = 128,
        activation: Module = ReLU(),
    ):
        layers = []
        for i in range(n_layers):
            layers.append(LazyLinear(n_units))
            layers.append(deepcopy(activation))
        layers.append(Linear(n_units, n_output))
        # layers.append(nn.Sigmoid())
        super().__init__(*layers)


class SingleWeight(Module):
    def __init__(
        self,
    ) -> None:
        super().__init__()
        self.weight = torch.nn.Parameter(torch.randn(()))

    def forward(self, x):
        prop1 = torch.ones_like(x[..., :1]) * self.weight
        prop2 = torch.ones_like(x[..., :1]) * (1 - self.weight)
        prop = torch.cat([prop1, prop2], dim=-1)
        return prop



from os import path as pt
from typing import List

import torch
from torch import Tensor
from torch.nn import Module

from neuralhedge.nn.datahedger import Hedger
from neuralhedge.nn.loss import log_utility


class Manager(Hedger):
    def __init__(self, strategy: Module, utility_func=log_utility):
        super().__init__(strategy)
        self.utility_func = utility_func

    def forward(self, prices: Tensor, info: Tensor):
        batch_size = prices.shape[0]
        wealth = [torch.ones([batch_size]) for t in range(prices.shape[1])]
        prop_hold = torch.zeros_like(prices)
        for t in range(prices.shape[1] - 1):
            info_dyn = wealth[t]
            all_info_t = self.compute_info_t(info_dyn, info, t)
            prop_hold[:, t + 1, :] = self.compute_prop_hold_tplus1(all_info_t, t)
            return_t = torch.sum(
                prop_hold[:, t + 1, :] * (prices[:, t + 1, :] / prices[:, t, :]),
                dim=-1,
                keepdim=False,
            )
            wealth[t + 1] = wealth[t] * return_t
        wealth = torch.cat([wealth_t.unsqueeze(1) for wealth_t in wealth], dim=1)
        return wealth

    def compute_prop_hold_tplus1(self, all_info_t: Tensor, t=None) -> Tensor:
        prop_hold_tplus1 = self.strategy(all_info_t)
        return prop_hold_tplus1

    def compute_info_t(self, info_dyn: Tensor, info: Tensor, t=None) -> Tensor:
        all_info_t = info[:, t, :]
        return all_info_t

    def compute_loss(self, input: List[Tensor]):
        prices, info = input
        terminal_wealth = self.forward(prices, info)[-1]
        return -self.utility_func(terminal_wealth)

    def record_history(self):
        return self.history["alpha"].append(list(self.parameters())[0].item())


class WealthManager(Manager):
    """
    Last coordinate of all_info_t is wealth
    """

    def __init__(self, model: Module, utility_func=...):
        super().__init__(model, utility_func)

    def compute_info_t(self, info_dyn: Tensor, info: Tensor, t=None) -> Tensor:
        info_dyn = info_dyn.unsqueeze(1)
        all_info_t = torch.cat([info[:, t, :], info_dyn], dim=-1)
        return all_info_t



from torch import Tensor
from torch.nn import Module
import torch.nn.functional as F


class EuropeanVanilla(Module):
    def __init__(
        self,
        strike: float,
        call: bool = True,
    ):
        super().__init__()
        self.call = call
        self.strike = strike

    def payoff(self, prices: Tensor) -> Tensor:
        """Returns the terminal payoff of financial derivative
        Shape: payoff: (n_sample, 1)
        """
        if self.call:
            payoff = F.relu(prices - self.strike)
        else:
            payoff = F.relu(self.strike - prices)
        return payoff



"""
Benchmark strategy for BlackScholes

"""

import torch
from torch.distributions.normal import Normal
from torch.nn import Module


class BlackScholesPrice(Module):
    def __init__(
        self,
        sigma,
        risk_free_rate,
        strike: float,
    ):
        super(BlackScholesPrice, self).__init__()
        self.sigma = torch.tensor(sigma)
        self.r = torch.tensor(risk_free_rate)
        self.strike = torch.tensor(strike)
        self.normal = Normal(0, 1)

    def forward(self, x):
        log_price = x[..., 0]
        time_to_maturity = x[..., 1]
        price = log_price.exp()
        d1 = (
            log_price
            - self.strike.log()
            + (self.r + self.sigma**2 / 2) * time_to_maturity
        ) / (self.sigma * time_to_maturity.sqrt())
        d2 = d1 - (self.sigma * time_to_maturity.sqrt())
        output = price * self.normal.cdf(d1) - self.strike * self.normal.cdf(
            d2
        ) * torch.exp(-self.r * time_to_maturity)
        return output[..., None]


class BlackScholesDelta(Module):
    def __init__(
        self,
        sigma,
        risk_free_rate,
        strike: float,
    ):
        super(BlackScholesDelta, self).__init__()
        self.sigma = torch.tensor(sigma)
        self.r = torch.tensor(risk_free_rate)
        self.strike = torch.tensor(strike)
        self.normal = Normal(0, 1)

    def forward(self, x):
        log_price = x[..., 0]
        time_to_maturity = x[..., 1]
        d1 = (
            log_price
            - self.strike.log()
            + (self.r + self.sigma**2 / 2) * time_to_maturity
        ) / (self.sigma * time_to_maturity.sqrt())
        return self.normal.cdf(d1)[..., None]


class BlackScholesAlpha(Module):
    def __init__(self, mu, sigma, r, alpha=None):
        super().__init__()
        self.mu = mu
        self.sigma = sigma
        self.r = r
        if alpha is None:
            self.alpha = (self.mu - self.r) / (self.sigma**2)
        else:
            self.alpha = alpha

    def forward(self, x):
        prop1 = torch.ones_like(x[..., :1]) * self.alpha
        prop2 = torch.ones_like(x[..., :1]) * (1 - self.alpha)
        prop = torch.cat([prop1, prop2], dim=-1)
        return prop


class BlackScholesMeanVarianceAlpha(BlackScholesAlpha):
    def __init__(self, mu, sigma, r, Wstar):
        super().__init__(mu, sigma, r)
        self.xi = (self.mu - self.r) / self.sigma
        self.Wstar = Wstar

    def compute_alpha(self, x):
        wealth = x[..., -1]
        time_to_maturity = x[..., 1]
        alpha = (
            self.xi
            / self.sigma
            / wealth
            * (self.Wstar * torch.exp(-self.r * time_to_maturity) - wealth)
        )
        return alpha.view(-1, 1)

    def forward(self, x):
        alpha = self.compute_alpha(x)
        prop = torch.cat([alpha, 1 - alpha], dim=-1)
        return prop


class BlackScholesMeanVarianceAlphaClip(BlackScholesMeanVarianceAlpha):
    def __init__(self, mu, sigma, r, Wstar, clip):
        super().__init__(mu, sigma, r, Wstar)
        self.clip = clip

    def forward(self, x):
        alpha = self.compute_alpha(x)

        alpha = alpha * (alpha > 0)
        alpha_clip = alpha * (alpha < self.clip) + self.clip * (alpha >= self.clip)

        prop = torch.cat([alpha_clip, 1 - alpha_clip], dim=-1)
        return prop



from typing import Union
import torch
from torch import Tensor
from torch.nn import Module
from torch.nn.parameter import Parameter
import torch.nn.functional as F


def proportional_cost(holding_diff, price_now) -> Tensor:
    cost = 0.001 * torch.abs(holding_diff) * price_now
    return cost


def admissible_cost(wealth, bound=0.0) -> Tensor:
    cost = F.relu(bound - torch.min(wealth, dim=1).values).mean()
    return cost


def no_cost(holding_diff, price_now) -> Tensor:
    cost = torch.tensor(0.0)
    return cost


def exp_utility(input: Tensor, a: float = 1.0) -> Tensor:
    """
    f(X) = -exp(-aX)
    """
    return -(-a * input).exp()


def log_utility(x: Tensor) -> Tensor:
    return x.log().mean()


def value_at_risk(input: Tensor, q: float = 0.01) -> Tensor:
    return torch.quantile(input, q, interpolation="linear")


def expected_shortfall(input: Tensor, q: float = 0.01) -> Tensor:
    VaR = value_at_risk(input, q)
    ES = F.relu(VaR - input).mean() / q - VaR
    return ES


class LossMeasure(Module):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)


class PowerMeasure(LossMeasure):
    @property
    def p(self):
        return self._p

    @p.setter
    def p(self, p: float = 1.0):
        if not p > 0:
            raise ValueError("Risk aversion coefficient should be positive.")
        self._p = p

    def __init__(self, p: float = 1.0) -> None:
        super().__init__()
        self.p = p

    def forward(self, input: Tensor) -> Tensor:
        """
        f(X) = (1/p) (E[max{X,0}^{p}])
        """
        input_T = input
        return (F.relu(-input_T) ** self.p).mean() ** (1 / self.p) / self.p


class EntropicRiskMeasure(LossMeasure):
    @property  # need property here?
    def a(self):
        return self._a

    @a.setter
    def a(self, a: float = 1.0):
        if not a > 0:
            raise ValueError("Risk aversion coefficient should be positive.")
        self._a = a

    def __init__(self, a: float = 1.0) -> None:
        super().__init__()
        self.a = a

    def forward(self, input_T: Tensor) -> Tensor:
        """
        rho(X) = (1/a) * log(E[exp(-aX)])
        """
        input_T_min = input_T.min()
        return (
            -exp_utility(input_T - input_T_min, a=self.a).mean()
        ).log() / self.a - input_T_min

    def optimal_omega(self, input_T: Tensor) -> Tensor:
        """
        input: (n_paths, n_steps+1, 1)
        f(X) = (1/a) * log(aE[exp(-aX)])
        """
        input_T_min = input_T.min()
        return (
            -exp_utility(input_T - input_T_min, a=self.a).mean() * self.a
        ).log() / self.a - input_T_min


class SquareMeasure(LossMeasure):
    @property
    def a(self):
        return self._a

    def __init__(self, a: float = 1.0) -> None:
        super().__init__()
        self._a = a

    def forward(self, input: Tensor) -> Tensor:
        """
        f(X) = Var(X)/2 - E[X]
        """
        input_T = input
        return input_T.var() / 2 - input_T.mean()

    def optimal_omega(self, input: Tensor) -> Tensor:
        """
        f(X) = -E[X]
        """
        input_T = input
        return -input_T.mean()


class ExpectedShortfall(LossMeasure):
    "q = 1-alpha"

    @property
    def q(self):
        return self._q

    @q.setter
    def q(self, q: float = 0.5):
        if not q > 0 or not q < 1:
            raise ValueError("Risk aversion coefficient should be between 0 and 1")
        self._q = q

    def __init__(self, q: float = 0.5) -> None:
        super().__init__()
        self._q = q

    def l_func(self, input: Tensor) -> Tensor:
        return F.relu(-input) / self.q

    def forward(self, input: Tensor) -> Tensor:
        """
        f(X) = ES_q(X)
        """
        input_T = input
        return expected_shortfall(input_T, self.q)

    def optimal_omega(self, input: Tensor) -> Tensor:
        """
        f(X) = -VaR_q(X)
        """
        input_T = input
        return -value_at_risk(input_T, self.q)
